module Schleuder
    # Wrapper for ruby-gpgme. Method naming is not strictly logical, this might
    # change but aliases will be set up then.
    class Crypt
      # Instantiates and stores password
      def initialize(password, options={})
        unless options[:new_keyring]
          # Check file permissions. ruby-gpgme unfortunately returns unspecific
          # errors if it can't read files.
          %w(pubring.gpg secring.gpg trustdb.gpg).each do |fn|
            f = File.join(ENV['GNUPGHOME'], fn)
            if ! File.readable?(f)
              raise Errno::EACCES.new('%s is not readable' % f)
            elsif ! File.writable?(f)
              raise Errno::EACCES.new('%s is not writable' % f)
            end
          end
        end

        @password = password
        if GPGME.respond_to? 'check_version'
          GPGME::check_version('0.0.0')
        end
        # Access to GPGME high-level API
        @crypto = GPGME::Crypto.new(:armor => true, :passphrase_callback => method(:passfunc))
        # Access to GPGME mid-level API
        @ctx = GPGME::Ctx.new(:armor => true, :passphrase_callback => method(:passfunc))
      end

      # Verify a gpg-signature. Use +signed_string+ if the signature is
      # detached. Returns a GPGME::SignatureResult
      def verify(sig, signed_string='')
        in_signed = ''
        if signed_string.empty?
          # verify +sig+ as cleartext (aka pgp/inline) signature
          Schleuder.log.debug 'No extra signed_string, verifying cleartext signature'
          output = @crypto.verify(sig) do |sig|
            in_signed = sig
          end
          output = output.read
        else
          # verify detached signature
          Schleuder.log.debug 'Verifying detached signature'
          @crypto.verify(sig, :signed_text => signed_string) do |sig|
            in_signed = sig
          end
          output = signed_string
          
        end
        Schleuder.log.debug 'verify_result: ' + in_signed.inspect
        
        [output, in_signed]
      end
      
      # Decrypt a string.
      def decrypt(str)
        output = ""
        in_encrypted = nil
        in_signed = nil
        
        # TODO: return ciphertext if missing key. Sensible e.g. if it is part
        # of a nested MIME-message and encrypted to someone else on purpose.
        # Breaking if even the whole message is not decryptable is a job for
        # the processor.
        
        # return input instead of empty String if not encrypted.
        unless str =~ /^-----BEGIN PGP MESSAGE-----/ || Magic.guess_string_mime_type(str).split('/').last.eql?('pgp')
          output, in_signed = verify(str)
          # match pgp-mime- and inline-pgp-signatures
          if str =~ /^-----BEGIN PGP SIG/
            Schleuder.log.debug 'found signed, not encrypted message, verifying'
            output, in_signed = verify(str)
          else
            Schleuder.log.debug 'found not signed, not encrypted message, returning input'
            output = str
          end
        else
          Schleuder.log.debug 'found pgp content, decrypting and verifying with gpgme'
          in_encrypted = true
          output = @crypto.decrypt(str, :passphrase_callback => method(:passfunc)) do |sig|
            in_signed = sig
          end
          output = output.read
          if output.empty?
            Exception.new("Output from GPGME::Crypto#decrypt was empty!")
          end
          # TODO: return mailadresses or keys instead of signature-objects?
        end
        [output, in_encrypted, in_signed]
      end
      
      # Encrypt a string to a single receiver and sign it. +receiver+ must be a
      # Schleuder::Member
      def encrypt_str(str, receiver)
        # encrypt and sign and return encrypted data as string
        # For some reason sometimes the last two characters of str are stolen
        # unless we append a blank or newline... Life is hard...
        @crypto.encrypt("#{str} ",
                        :recipients => [receiver.key],
                        :sign => true,
                        :always_trust => true
                       ).read
      end
      
      # Lists all public keys matching +pattern+. Returns an array of
      # GPGME::GpgKey's
      def list_keys(pattern='')
        @ctx.keys(pattern)
      end

      # Returns the GPGME::GpgKey matching +pattern+. Log an error if more than
      # one matches, because duplicated user-ids is a sensitive issue.
      def get_key(pattern, only_valid_keys=false)
        pattern = "<#{pattern}>" if pattern =~ /.*@.*/ && !(pattern =~ /^<.*>$/)
        keys = list_keys(pattern)

        keys.reject! { |key| [:revoked,:expired].include?(key.trust) } if only_valid_keys

        if keys.empty?
          [false, "no key found for #{pattern}."]
        elsif keys.length > 1
          Schleuder.log.warn "There's more than one key matching the pattern you gave me!"
          Schleuder.log.debug { "Pattern: #{pattern.inspect}" }
          Schleuder.log.debug { "Keys: #{keys.inspect}" }
          [false, "no distinct key for #{pattern.inspect} found. Matching keys: #{key_infos(keys,only_valid_keys).join(', ')}"]
        else
          [keys.first]
        end
      end
      
      # Signs +string+ with the private key of the list (aka detached signature)
      def sign(string)
        @crypto.detach_sign(string).read
      end

      # Clearsigns +string+ with the private key of the list
      def clearsign(string)
        @crypto.clearsign(string).read
      end
      
      # Exports the public key matching +keyid+ as ascii key block.
      def export(keyid)
        # using #read on the resulting GPGME::Data yields an empty string... weird
        @ctx.export_keys(keyid).inspect
      end
      
      # Delete the public key matching +pattern+ from the public key ring of the
      # list
      def delete_key(key)
        msg = nil
        key, msg = get_key(key) if key.kind_of?(String)

        if key
          @ctx.delete_key(key)
          return true
        else
          msg
        end
      rescue => e
        return e
      end
      
      # Import +keydata+ into public key ring of the list
      def add_key(keydata)
        @ctx.import_keys(GPGME::Data.new(keydata))
        @ctx.import_result
      end

      def add_key_from_file(keyfile)
        add_key(File.read(keyfile))
      end

      def key_descr(key)
        key.to_s.split("\n").first.split[1..2].join(' ')
      end

      private

      def key_infos(keys, only_valid_keys=false)
        keys.collect { |key| key_descr(key) + (!only_valid_keys && [:revoked, :expired].include?(key.trust)) ? " *#{key.trust}*" : '' }
      end

      def passfunc(hook, uid_hint, passphrase_info, prev_was_bad, fd)
        io = IO.for_fd(fd, 'w')
        io.puts @password
        io.flush
      end

    end
end
